/****************************************************************************
*                                                                           *
*   SplitMessages module for Kadu                                           *
*   Copyright (C) 2011  Piotr Dąbrowski ultr@ultr.pl                        *
*                                                                           *
*   This program is free software: you can redistribute it and/or modify    *
*   it under the terms of the GNU General Public License as published by    *
*   the Free Software Foundation, either version 3 of the License, or       *
*   (at your option) any later version.                                     *
*                                                                           *
*   This program is distributed in the hope that it will be useful,         *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of          *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           *
*   GNU General Public License for more details.                            *
*                                                                           *
*   You should have received a copy of the GNU General Public License       *
*   along with this program.  If not, see <http://www.gnu.org/licenses/>.   *
*                                                                           *
****************************************************************************/




#include <QApplication>
#include <QTextDocumentFragment>

#include "configuration/configuration-file.h"
#include "gui/widgets/chat-edit-box.h"
#include "gui/widgets/chat-widget-manager.h"
#include "gui/widgets/custom-input.h"
#include "misc/misc.h"
#include "debug.h"

#include "../encryption_ng/encryption-chat-data.h"

#include "splitmessages.h"

#include "defines.h"




SplitMessages *splitmessages;




extern "C" int splitmessages_init()
{
	kdebugf();
	splitmessages = new SplitMessages();
	MainConfigurationWindow::registerUiFile( dataPath("kadu/modules/configuration/splitmessages.ui") );
	kdebugf2();
	return 0;
}


extern "C" void splitmessages_close()
{
	kdebugf();
	MainConfigurationWindow::unregisterUiFile( dataPath("kadu/modules/configuration/splitmessages.ui") );
	delete splitmessages;
	splitmessages = NULL;
	kdebugf2();
}


SplitMessages::SplitMessages() : QObject()
{
	partssendinglock = false;
	// configuration handling
	createDefaultConfiguration();
	// read the configuration and force its usage
	configurationUpdated();
	// connect chat widgets events and handle opened ones
	connect( ChatWidgetManager::instance(), SIGNAL(chatWidgetCreated(ChatWidget*))   , this, SLOT(chatCreated(ChatWidget*))    );
	connect( ChatWidgetManager::instance(), SIGNAL(chatWidgetDestroying(ChatWidget*)), this, SLOT(chatDestroying(ChatWidget*)) );
	foreach( ChatWidget *chatwidget, ChatWidgetManager::instance()->chats() )
		chatCreated( chatwidget );
}


SplitMessages::~SplitMessages()
{
	// disconnect chat widgets events
	disconnect( ChatWidgetManager::instance(), SIGNAL(chatWidgetCreated(ChatWidget*))   , this, SLOT(chatCreated(ChatWidget*))    );
	disconnect( ChatWidgetManager::instance(), SIGNAL(chatWidgetDestroying(ChatWidget*)), this, SLOT(chatDestroying(ChatWidget*)) );
}


void SplitMessages::mainConfigurationWindowCreated( MainConfigurationWindow *mainConfigurationWindow )
{
	Q_UNUSED( mainConfigurationWindow );
}


void SplitMessages::createDefaultConfiguration()
{
	config_file.addVariable( "SplitMessages", "SmartMessagesSplitting", true                        );
	config_file.addVariable( "SplitMessages", "Delay"                 , SPLITMESSAGES_DEFAULT_DELAY );
}


void SplitMessages::configurationUpdated()
{
}


void SplitMessages::chatCreated( ChatWidget *chatwidget )
{
	connect( chatwidget, SIGNAL( messageSendRequested( ChatWidget* ) ), this, SLOT( messageSendRequested( ChatWidget* ) ) );
}


void SplitMessages::chatDestroying( ChatWidget *chatwidget )
{
	disconnect( chatwidget, SIGNAL( messageSendRequested( ChatWidget* ) ), this, SLOT( messageSendRequested( ChatWidget* ) ) );
}


void SplitMessages::messageSendRequested( ChatWidget *chatwidget )
{
	// check parts sending
	if( partssendinglock )
		return;
	// check protocol
	if( chatwidget->chat().chatAccount().protocolName() != "gadu" )
		return;
	// data
	int maxlength = SPLITMESSAGES_MAXIMUM_LENGTH_NORMAL;
	int maxbackward = SPLITMESSAGES_MAXIMUM_BACKWARD;
	// check encryption
	EncryptionChatData *encryptionChatData = chatwidget->chat().data()->moduleStorableData<EncryptionChatData>("encryption-ng", this, true);
	if( encryptionChatData && encryptionChatData->encryptor() )
		maxlength = SPLITMESSAGES_MAXIMUM_LENGTH_ENCRYPTED;
	// start
	CustomInput *edit = chatwidget->edit();
	edit->setUpdatesEnabled( false );
	QList<QTextDocumentFragment> messages;
	while( ! edit->toPlainText().isEmpty() )
	{
		// move cursor to maxlength
		edit->moveCursor( QTextCursor::Start );
		while( true )
		{
			edit->moveCursor( QTextCursor::Right, QTextCursor::KeepAnchor );
			int l = edit->textCursor().selectedText().toUtf8().length();
			if( ( l >= maxlength ) || edit->textCursor().atEnd() )
			{
				if( l > maxlength )
					edit->moveCursor( QTextCursor::Left, QTextCursor::KeepAnchor );
				break;
			}
		}
		QTextCursor tc;
		tc = edit->textCursor();
		tc.clearSelection();
		edit->setTextCursor( tc );
		// smart splitting
		if( ( ! edit->textCursor().atEnd() ) && config_file.readBoolEntry( "SplitMessages", "SmartMessagesSplitting", false ) )
		{
			while( true )
			{
				int position = edit->textCursor().position();
				if( position <= maxbackward )
					break;
				QTextCursor tc;
				// block
				edit->moveCursor( QTextCursor::StartOfBlock );
				if( ( edit->textCursor().position() > 0 ) && ( position - edit->textCursor().position() < maxbackward ) )
				{
					edit->textCursor().deletePreviousChar();
					break;
				}
				tc = edit->textCursor();
				tc.setPosition( position );
				edit->setTextCursor( tc );
				// line
				edit->moveCursor( QTextCursor::StartOfLine );
				if( ( edit->textCursor().position() > 0 ) && ( position - edit->textCursor().position() < maxbackward ) )
				{
					edit->textCursor().deletePreviousChar();
					break;
				}
				tc = edit->textCursor();
				tc.setPosition( position );
				edit->setTextCursor( tc );
				// word
				edit->moveCursor( QTextCursor::StartOfWord );
				if( ( edit->textCursor().position() > 0 ) && ( position - edit->textCursor().position() < maxbackward ) )
				{
					break;
				}
				tc = edit->textCursor();
				tc.setPosition( position );
				edit->setTextCursor( tc );
				break;
			}
		}
		// select part, save it and remove
		edit->moveCursor( QTextCursor::Start, QTextCursor::KeepAnchor );
		messages.append( edit->textCursor().selection() );
		edit->textCursor().removeSelectedText();
	}
	// send
	if( messages.count() == 1 )
	{
		edit->clear();
		edit->textCursor().insertFragment( messages[0] );
	}
	else
	{
		partssendinglock = true;
		for( int k = 0; k < messages.count(); k++ )
		{
			edit->clear();
			edit->textCursor().insertFragment( messages[k] );
			if( k < messages.count() - 1 )
			{
				chatwidget->sendMessage();
				wait( config_file.readNumEntry( "SplitMessages", "Delay", SPLITMESSAGES_DEFAULT_DELAY ) );
			}
		}
		partssendinglock = false;
	}
	// done
	edit->setUpdatesEnabled( true );
}


void SplitMessages::wait( int milliseconds )
{
	int t = 100;
	do
	{
		usleep( qMin( t, milliseconds ) * 1000 );
		qApp->blockSignals( true );
		qApp->processEvents( QEventLoop::ExcludeUserInputEvents );
		qApp->blockSignals( false );
		milliseconds -= t;
	} while( milliseconds > 0 );
}
